/********************************************************************************
 * 
 * CMTF: Coupled Matrix Tensor Factorization
 * Code by Inah Jeon and U Kang
 * 
 * ------------------------------------------------------------------------------
 * 
 * File: CMTF.java
 * - It performs coupled matrix tensor Factorization
 * - Outputs are stored in <output_path> on HDFS
 * 
 * ------------------------------------------------------------------------------
 * 
 * Version: 1.0
 * 
 *********************************************************************************/

package SCouT;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Date;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.conf.Configured;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.FileUtil;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapred.FileInputFormat;
import org.apache.hadoop.mapred.FileOutputFormat;
import org.apache.hadoop.mapred.JobClient;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.TextInputFormat;
import org.apache.hadoop.mapred.lib.MultipleInputs;
import org.apache.hadoop.util.Tool;
import org.apache.hadoop.util.ToolRunner;

import Jama.Matrix;
import Jama.SingularValueDecomposition;
import SCouT.Collapse.CollapseMultipleModeMapper;
import SCouT.ConvertRowwise.ConvertRowwiseMapper;
import SCouT.ConvertRowwise.ConvertRowwiseReducer;
import SCouT.InitFactor.InitFactorMapper;
import SCouT.MTM.MTMMapper;
import SCouT.MTM.MTMReducer;
import SCouT.NModeMatrixHadamardProduct.NModeMatrixHadamardProductFactorMapper;
import SCouT.NModeMatrixHadamardProduct.NModeMatrixHadamardProductTensorMapper;
import SCouT.SuccessiveNModeMatrixHadamardProduct.SuccessiveNModeMatrixHadamardProductFactorMapper;
import SCouT.SuccessiveNModeMatrixHadamardProduct.SuccessiveNModeMatrixHadamardProductReducer;
import SCouT.SuccessiveNModeMatrixHadamardProduct.SuccessiveNModeMatrixHadamardProductTensorMapper;


public class CMTFtest extends Configured implements Tool {
	int iteration = 0;	// check current iteration

	// parameters for CMTF
	int rank; // R for rank
	int nreducers, niter; // nreducers: # reducer, niter: # iteration of PARAFAC
	ArrayList<Long> dimensionsTensor = new ArrayList<Long>(); // dimensionalities of input tensor
	ArrayList<Integer> coupledMode = new ArrayList<Integer>(); // coupled modes
	ArrayList<Long> dimensionsMatrix = new ArrayList<Long>(); // dimensionalities of coupled matrices

	// input, output paths
	String in_tensor_path;	// input tensor path
	ArrayList<String> in_matrix_path = new ArrayList<String>(); // input matrix path
	String out_path;	// output path of CMTF
	String out_tensor_factor_path = null;
	String out_matrix_factor_path = null;
	Path temp_path = new Path("temp_cmtf");

	String pseudoinv_str; // pseudo inverse values string
	double[] lambda = null; // lambda values
	FileSystem fs; // HDFS

	int init;
	String log_name;

	// Main entry point.
	public static void main (final String[] args) throws Exception {
		final int result = ToolRunner.run(new Configuration(), new CMTFtest(), args);

		System.exit(result);
	}

	// Print the command-line usage text.
	protected static int printUsage () {
		System.out.println("BigTensor:CMTF <dim_1:dim_2:...:dim_N (dimensions of tensor)> <mode_1:mode_2:...:mode_N (coupled modes)> <dim_1:dim_2:...:dim_N (dimensions of matrix)> <Rank> <# of reducer> <max iteration> <tensor_path> <path_1:path2:...:path_N (paths of coupled matrices)> <output_path>");
		ToolRunner.printGenericCommandUsage(System.out);

		return -1;
	}

	public int run(String[] args) throws Exception {
		if( args.length != 11 ) {
			System.out.println("len: "+args.length);
			return printUsage();
		}

		// set parameters and clean up HDFS paths, and files
		fs = FileSystem.get(getConf());
		setParameters(args);
		cleanupFiles();

		System.out.println("\n-----===[BigTensor: A Scalable Tensor Decomposition System]===-----\n");
		System.out.println("[BigTensor:CMTF] Computing coupled matrix tensor Factorization. #mode= "+ dimensionsTensor.size() + ", Rank=" + rank + ", reducers=" + nreducers + ", max iteration=" + niter + "\n");

		// initialize factor matrices
		if(init == 0) prepareInitialFactors();
		FileUtil.copy(fs, new Path(out_path), fs, new Path(out_path+"_iter0"), false, getConf());

		// Start time
		Date startTime = new Date();

		// Main loop
		for (int iter=1; iter<=niter; iter++) {
			iteration = iter;
			System.out.println("[BigTensor:CMTF] Iteration: " +iteration+ "\n");

			// Start time
			Date substartTime = new Date();
			// update tensor factors
			for(int m=1; m<=1; m++) {
				if(coupledMode.contains(m)) {
					updateTensorFactor(m, true);	
				} else {
					updateTensorFactor(m, false);	
				}
			}
			// End time
			Date subendTime = new Date();
			subTimeCheck(substartTime, subendTime, "TF_" + log_name);	

			// update matrix factors
			if(args[1].charAt(0) != '0') {
				for(int m=1; m<=coupledMode.size(); m++) {
					updateMatrixFactor(coupledMode.get(m-1));
				}
			}
			//FileUtil.copy(fs, new Path(out_path), fs, new Path(out_path+"_iter"+iter), false, getConf());
		}

		// End time
		Date endTime = new Date();

		// print running time
		subTimeCheck(startTime, endTime, "Total_"+log_name);	

		return 0;
	}

	/**
	 * 
	 * setParameters: Setting parameters of PARAFAC by input arguments
	 * - input: args[] - input arguments  
	 * 	args[0]: dim_1:dim_2:...:dim_N (dimensions of tensor)
	 * 	args[1]: mode_1:mode_2:...:mode_N (coupled modes)
	 * 	args[2]: dim_1:dim_2:...:dim_N (dimensions of matrix)
	 * 	args[3]: rank of CMTF
	 * 	args[4]: # of reducer
	 * 	args[5]: max iteration
	 * 	args[6]: tensor_path
	 * 	args[7]: path_1:path2:...:path_N (paths of coupled matrices) 
	 * 	args[8]: output_path
	 * 	args[9]: initialization
	 * 	args[10]: log_name
	 * - output: none
	 * 
	 */
	private void setParameters(String[] args) {
		// --- parameters for CMTF 

		// dimensions of tensor
		String[] tokens = args[0].split(":");
		for(int i=0; i<tokens.length; i++) {
			dimensionsTensor.add(Long.parseLong(tokens[i]));
		}

		// coupled modes
		if(args[1].charAt(0) != '0') {
			tokens = args[1].split(":");
			for(int i=0; i<tokens.length; i++) {
				coupledMode.add(Integer.parseInt(tokens[i]));
			}

			//  dimensionalities of coupled matrices
			tokens = args[2].split(":");
			for(int i=0; i<tokens.length; i++) {
				dimensionsMatrix.add(Long.parseLong(tokens[i]));
			}

			// input matrix path
			tokens = args[7].split(":");
			for(int i=0; i<tokens.length; i++) {
				in_matrix_path.add(tokens[i]);
			}
		}

		// R for rank of CMTF
		rank = Integer.parseInt(args[3]);
		// nreducers: # reducer
		nreducers = Integer.parseInt(args[4]);
		// niter: # iteration of CMTF
		niter = Integer.parseInt(args[5]);

		// input tensor path
		in_tensor_path = args[6];

		out_path = args[8];	// output path of CMTF
		out_tensor_factor_path = out_path + "/tensor_factor";
		out_matrix_factor_path = out_path + "/matrix_factor";

		init = Integer.parseInt(args[9]);
		log_name = args[10];
	}

	// Cleaning up HDFS files
	// input: none
	// output: none
	private void cleanupFiles() throws IOException {
		fs.delete(temp_path, true);
		fs.mkdirs(temp_path);
		if(init == 0) {
			fs.delete(new Path(out_path), true);
			fs.mkdirs(new Path(out_path));
		}
	}

	// Initialization of factors
	// input: none
	// output: initialized factors
	private void prepareInitialFactors() throws IOException, Exception {
		// Initialize factor matrices with random values
		for(int i=1; i<=dimensionsTensor.size(); i++) {
			JobClient.runJob(configInitFactor (dimensionsTensor.get(i-1), i, new Path(out_tensor_factor_path + "/" + i)));
		}
		for(int i=1; i<=dimensionsMatrix.size(); i++) {
			JobClient.runJob(configInitFactor (dimensionsMatrix.get(i-1), coupledMode.get(i-1), new Path(out_matrix_factor_path + "/" + coupledMode.get(i-1))));
		}
		//fs.copyFromLocalFile(new Path(out_path), new Path("initialized_factor"));
		fs.delete(new Path("init_temp"), true);
	}

	// Update tensor factor
	// input: none
	// output: update tensor factor
	private void updateTensorFactor(int current_mode, boolean is_coupled) throws Exception, IOException {
		// --- Update tensor factor
		System.out.println("[BigTensor:CMTF]" +iteration+ "th updating Coupled Tensor factor (mode:"+current_mode+")");

		Path MTM_out_path = new Path(temp_path + "/MTM_out");
		String rowwise_tensor_factor = temp_path + "/temp_tensor";
		String rowwise_matrix_factor = temp_path + "/temp_matrix";

		// 1. CTC = C^T * C, BTB = B^T * B	
		int cmode = 1;
		if(current_mode != 1) {
			JobClient.runJob(configConvertRowwise (out_tensor_factor_path + "/" + 1, new Path(rowwise_tensor_factor + 1), 1));
			JobClient.runJob( configMTM(new Path(rowwise_tensor_factor + 1), MTM_out_path) );
			cmode ++;
		} else {
			JobClient.runJob(configConvertRowwise (out_tensor_factor_path + "/" + 2, new Path(rowwise_tensor_factor + 2), 2));
			JobClient.runJob( configMTM(new Path(rowwise_tensor_factor + 2), MTM_out_path) );
			cmode += 2;
		}
		Matrix tfactorTtfactor = merge_read_matrix(MTM_out_path, rank, rank);


		for(int i=cmode; i<=dimensionsTensor.size(); i++) {
			if(i != current_mode) {
				System.out.println("[BigTensor:CMTF]" +iteration+ "th Computing tfactorTtfactor = tfactor^T * tfactor	\n");
				JobClient.runJob(configConvertRowwise (out_tensor_factor_path + "/" + i, new Path(rowwise_tensor_factor + i), i));
				Date MTMStartTime = new Date();
				JobClient.runJob( configMTM(new Path(rowwise_tensor_factor + i), MTM_out_path));
				// End time
				Date MTMEndTime = new Date();

				// print running time
				subTimeCheck(MTMStartTime, MTMEndTime, "MTM_" + log_name);

				tfactorTtfactor = tfactorTtfactor.arrayTimes(merge_read_matrix(MTM_out_path, rank, rank));
			}
		}

		if(is_coupled == true) {
			// 2. DTD = D^T * D	
			System.out.println("[BigTensor:CMTF]" +iteration+ "th Computing mfactorTmfactor = mfactor^T * mfactor\n");
			JobClient.runJob(configConvertRowwise (out_matrix_factor_path + "/" + current_mode, new Path(rowwise_matrix_factor + current_mode), current_mode));
			JobClient.runJob( configMTM(new Path(rowwise_matrix_factor + current_mode), MTM_out_path) );
			tfactorTtfactor = tfactorTtfactor.plus(merge_read_matrix(MTM_out_path, rank, rank));
		}

		ArrayList<Path> temp_MHP_path = new ArrayList<Path>();
		cmode = 0;
		Path temp_nMMHP_prev_path = new Path(temp_path + "/temp_nMMHP" + current_mode + "prev");
		Path temp_nMMHP_next_path = new Path(temp_path + "/temp_nMMHP" + current_mode + "next");

		// Start time
		Date startTime = new Date();

		while(cmode < dimensionsTensor.size()) {
			if(current_mode != cmode+1) {
				if(cmode == 0 || (cmode == 1 && current_mode == 1)) {
					System.out.println("[BigTensor:CMTF]" +iteration+ "th Computing MHP1 cmode:" + cmode + " temp_nMMHP_prev_path:" + temp_nMMHP_prev_path.toString() + " temp_nMMHP_next_path:" + temp_nMMHP_next_path.toString() + " \n");
					JobClient.runJob(configNModeMatrixHadamardProduct(new Path(in_tensor_path), new Path(rowwise_tensor_factor + (cmode+1)), cmode+1, temp_nMMHP_next_path));
				} else {
					System.out.println("[BigTensor:CMTF]" +iteration+ "th Computing MHP2 cmode:" + cmode + " temp_nMMHP_prev_path:" + temp_nMMHP_prev_path.toString() + " temp_nMMHP_next_path:" + temp_nMMHP_next_path.toString() + " \n");
					JobClient.runJob(configSuccessiveNModeMatrixHadamardProduct(temp_nMMHP_prev_path, new Path(rowwise_tensor_factor + (cmode+1)), cmode+1, temp_nMMHP_next_path));
				}
				Path temp = temp_nMMHP_prev_path;
				temp_nMMHP_prev_path = temp_nMMHP_next_path;
				temp_nMMHP_next_path = temp;
			}
			cmode++;
		}	

		temp_MHP_path.add(temp_nMMHP_prev_path);


		if(is_coupled == true) {
			String matrix_path = in_matrix_path.get(coupledMode.indexOf(current_mode));
			// 5. YD = Y * ((CTC * BTB + DTD)^\dagger * D^T)^T
			temp_MHP_path.add(new Path(temp_path + "/temp_MHP" + current_mode));
			JobClient.runJob(configNModeMatrixHadamardProduct(new Path(matrix_path), new Path(rowwise_matrix_factor + current_mode), 2, new Path(temp_path + "/temp_MHP" + current_mode)));
		}

		// 6. collapse (XCB, YD) into current_mode
		Path collapsed_path = new Path(temp_path + "/collapsed_MSHP");
		JobClient.runJob(configMultipleCollapse(temp_MHP_path, current_mode, collapsed_path));

		// End time
		Date endTime = new Date();

		// print running time
		subTimeCheck(startTime, endTime, "MMHP_" + log_name);

		// 3. (CTC * BTB + DTD)^\dagger
		String pinv_tfactorTtfactor = getPseudoInverseString(tfactorTtfactor);

		JobClient.runJob(configMatrixMatrixStrProduct(collapsed_path, current_mode, pinv_tfactorTtfactor, new Path(out_tensor_factor_path + "/" + current_mode)));
		fs.delete(temp_path, true);
		fs.mkdirs(temp_path);
	}

	// Update matrix factor
	// input: none
	// output: update matrix factor
	private void updateMatrixFactor(int current_mode) throws Exception, IOException {
		// --- Update matrix factor
		Path temp_MHP_path = new Path(temp_path + "/temp_MHP" + current_mode);
		String rowwise_tensor_factor = temp_path + "/temp_tensor";
		String matrix_path = in_matrix_path.get(coupledMode.indexOf(current_mode));
		Path MTM_out_path = new Path(temp_path + "/MTM_out");

		// pinvA^T = (A^/dagger)^T = A * pinv(A^T A)
		JobClient.runJob(configConvertRowwise (out_tensor_factor_path + "/" + current_mode, new Path(rowwise_tensor_factor + current_mode), current_mode));
		JobClient.runJob( configMTM(new Path(rowwise_tensor_factor + current_mode), MTM_out_path) );
		Matrix tfactorTtfactor = merge_read_matrix(MTM_out_path, rank, rank);
		String pinv_tfactorTtfactor = getPseudoInverseString(tfactorTtfactor);

		// A * pinv(A^TA)
		JobClient.runJob(configMatrixMatrixStrProduct(new Path(rowwise_tensor_factor + current_mode), current_mode, pinv_tfactorTtfactor, new Path(temp_path + "/pinvAT")));
		JobClient.runJob(configConvertRowwise ((temp_path+"/pinvAT"), new Path(rowwise_tensor_factor + current_mode + "inv"), current_mode));

		// Y transpose
		JobClient.runJob(configMatrixTranspose( new Path(matrix_path), new Path(temp_path + "/YT")));

		// D = Y1^T * pinvA^T
		JobClient.runJob(configNModeMatrixHadamardProduct(new Path(temp_path+"/YT"), new Path(rowwise_tensor_factor + current_mode + "inv"), 2, temp_MHP_path));
		JobClient.runJob(configCollapse(temp_MHP_path, 1, new Path(temp_path + "/collapsed_MSHP")));
		JobClient.runJob(configConvertColumnwise(new Path(temp_path + "/collapsed_MSHP"), new Path(out_matrix_factor_path + "/" + current_mode), current_mode));

		fs.delete(temp_path, true);
		fs.mkdirs(temp_path);
	}

	/**
	 * 
	 * getPseudoInverseString: get pseudo inverse of a matrix
	 * - input:
	 * 	mat: input matrix
	 * - output:
	 * 	string of pseudo inverse of input matrix
	 * 
	 */
	private String getPseudoInverseString(Matrix mat) {
		String pinv_mat = "";
		for(int i = 0; i < rank; i++) {
			for(int j = 0; j < rank; j++) {
				pinv_mat += "" + mat.get(i, j) + "\t";
			}
		}
		System.out.println("[Debug getPseudoInverseString] "+ pinv_mat);

		SingularValueDecomposition svd = new SingularValueDecomposition( mat );

		Matrix U = svd.getU();
		Matrix S = svd.getS();
		Matrix V = svd.getV();
		Matrix inv_mat_temp = V.times( S.inverse() );
		Matrix inv_mat = inv_mat_temp.times( U.transpose() );
		pinv_mat = "";
		double[][] inv_array = inv_mat.getArray();
		for(int i = 0; i < rank; i++) {
			for(int j = 0; j < rank; j++) {
				pinv_mat += "" + inv_array[i][j] + "\t";
			}
		}
		return pinv_mat;
	}

	/**
	 * 
	 * read_matrix: Read matrix values from file and return it in matrix array
	 * - input:
	 * 	in_path: input path to read matrix
	 * 	rowdim, coldim: row and column dimensionalities of matrix to read
	 * - output:
	 * 	read matrix from the input path
	 * 
	 */
	public Matrix merge_read_matrix(Path in_path, int rowdim, int coldim) throws IOException {
		FileSystem fs = FileSystem.get(getConf());
		FileSystem lfs = FileSystem.getLocal(getConf());
		Configuration conf = new Configuration();
		String local_output_path = "temp_matrix";
		lfs.delete(new Path(local_output_path), true);

		Matrix matrix = new Matrix(rowdim,coldim);
		for(int i =0; i < rowdim; i++) {
			for(int j=0; j< coldim; j++) {
				matrix.set(i, j, 0);
			}
		}

		// read the result
		FileUtil.copyMerge(fs, in_path, lfs, new Path(local_output_path), false, conf, null);
		BufferedReader br=new BufferedReader(new InputStreamReader(new FileInputStream( local_output_path ), "UTF8"));
		String line = "";

		while((line = br.readLine()) != null) {
			System.out.println("merge read_matrix "+line); // 0 val val val ...
			String[] tokens = line.split("\t");
			int row = 0;
			int col = 0;
			for(int i=1; i<tokens.length; i++) {
				row = (i-1)/coldim;
				col = (i-1)%coldim;
				matrix.set(row, col, matrix.get(row, col) + Double.parseDouble(tokens[i]));
			}
		}

		br.close();
		lfs.delete(new Path(local_output_path), true);
		return matrix;
	}

	/**
	 * 
	 * read_matrix: Read matrix values from file and return it in matrix array
	 * - input:
	 * 	in_path: input path to read matrix
	 * 	rowdim, coldim: row and column dimensionalities of matrix to read
	 * - output:
	 * 	read matrix from the input path
	 * 
	 */
	public Matrix read_matrix(Path in_path, int rowdim, int coldim) throws IOException {
		Matrix matrix = new Matrix(rowdim,coldim);
		FileSystem fs = FileSystem.get(getConf());
		FileSystem lfs = FileSystem.getLocal(getConf());
		Configuration conf = new Configuration();
		String local_output_path = "temp_matrix";
		lfs.delete(new Path(local_output_path), true);

		// read the result
		FileUtil.copyMerge(fs, in_path, lfs, new Path(local_output_path), false, conf, null);
		BufferedReader br=new BufferedReader(new InputStreamReader(new FileInputStream( local_output_path ), "UTF8"));
		String line;
		line=br.readLine();
		System.out.println("read_matrix "+line); // 0 val val val ...
		String[] tokens = line.split("\t");
		int row = 0;
		int col = 0;
		for(int i=1; i<tokens.length; i++) {
			row = (i-1)/coldim;
			col = (i-1)%coldim;
			matrix.set(row, col, Double.parseDouble(tokens[i]));
		}

		br.close();
		lfs.delete(new Path(local_output_path), true);
		return matrix;
	}

	// Check running time and write log
	// input:
	// 	- startTime: Start time of entire program 
	// 	- endTime: End time of entire program
	// output: none
	private void subTimeCheck(Date startTime, Date endTime, String logfile) throws IOException {
		// Time Check
		long lTime = endTime.getTime() - startTime.getTime();
		System.out.println("TIME : " + lTime + "(ms)");

		String tmp = "Start:" + startTime.toString() + " End: " +endTime.toString()+ " Running Time: " + lTime + "\n";
		char intext2[] = new char[tmp.length()];
		tmp.getChars(0, tmp.length(), intext2, 0);
		final File someFile2 = new File("./", logfile);
		someFile2.createNewFile();
		FileWriter fw2 = new FileWriter(logfile);
		fw2.write(intext2);
		fw2.close();
		System.out.println(tmp);
	}

	// Check running time and write log
	// input:
	// 	- startTime: Start time of entire program 
	// 	- endTime: End time of entire program
	// output: none
	private void timeCheck(Date startTime, Date endTime) throws IOException {
		// Time Check
		long lTime = endTime.getTime() - startTime.getTime();
		System.out.println("TIME : " + lTime + "(ms)");

		String tmp = "Start:" + startTime.toString() + " End: " +endTime.toString()+ " Running Time: " + lTime + "\n";
		char intext2[] = new char[tmp.length()];
		tmp.getChars(0, tmp.length(), intext2, 0);
		final File someFile2 = new File("./", log_name);
		someFile2.createNewFile();
		FileWriter fw2 = new FileWriter(log_name);
		fw2.write(intext2);
		fw2.close();
		System.out.println(tmp);
	}

	/**
	 * 
	 * configInitFactor: Configure initialize (make factor matrices)
	 * - input:
	 * 	current_mode: current_mode of factor
	 * 	current_column: current column of factor
	 * - output:
	 * 	a column of initialized factor matrix
	 * 
	 */
	protected JobConf configInitFactor (long dim, int current_mode, Path out_path) throws Exception {
		final JobConf conf = new JobConf(getConf(), CMTFtest.class);
		conf.set("dim", "" + dim);
		conf.set("current_mode", "" + current_mode);
		conf.set("rank", "" + rank);
		conf.set("nreducers", "" + nreducers);
		conf.setJobName("[Iteration "+ iteration +"] InitFactor (mode: " +current_mode+ ")");

		conf.setMapperClass(InitFactorMapper.class);        

		fs.delete(out_path, true);
		fs.delete(new Path("init_temp"), true);

		// Initialize paths of factor matrices
		fs.mkdirs(new Path("init_temp"));
		for(int i=0; i<nreducers; i++) {
			String tmp = "" + i;
			Path outFile = new Path("init_temp/in"+i);
			BufferedWriter br=new BufferedWriter(new OutputStreamWriter(fs.create(outFile,true)));
			br.write(tmp);
			br.close();
		}

		FileInputFormat.setInputPaths(conf, new Path("init_temp"));  
		FileOutputFormat.setOutputPath(conf, out_path);  
		conf.setNumReduceTasks( 0 );

		conf.setOutputKeyClass(IntWritable.class);
		conf.setOutputValueClass(Text.class);
		return conf;
	}

	// Configure MatrixMatrixStrProduct
	// input:
	// 	- current_mode: current mode
	// 	- out_path: output path
	// output: result of n-mode MatrixHadamardProduct
	protected JobConf configMatrixMatrixStrProduct (Path input_path, int current_mode, String matrix_str, Path output_path) throws Exception {
		final JobConf conf = new JobConf(getConf(), CMTFtest.class);
		conf.set("rowdim", "" + rank);
		conf.set("coldim", "" + rank);
		conf.set("current_mode", "" + current_mode);
		conf.set("matrix_str", "" + matrix_str);
		conf.setJobName("[Iteration "+ iteration +"] current_mode="+current_mode+" MatrixMatrixStrProduct");

		conf.setMapperClass(MatrixMatrixStrProduct.MatrixMatrixStrProductMapper.class);                

		fs.delete(output_path, true);

		FileInputFormat.setInputPaths(conf, input_path);
		FileOutputFormat.setOutputPath(conf, output_path);  

		conf.setNumReduceTasks( 0 );

		conf.setOutputKeyClass(IntWritable.class);
		conf.setOutputValueClass(Text.class);

		return conf;
	}

	// Configure MatrixHadamardProduct
	// input:
	// 	- current_mode: current mode
	// 	- out_path: output path
	// output: result of n-mode MatrixHadamardProduct
	protected JobConf configNModeMatrixHadamardProduct (Path input_tensor_path, Path input_factor_path, int product_mode, Path output_path) throws Exception {
		final JobConf conf = new JobConf(getConf(), CMTFtest.class);

		conf.set("product_mode", "" + product_mode);
		conf.set("rank", "" + rank);
		conf.setJobName("[Iteration "+ iteration +"] product_mode="+product_mode+" NModeMatrixHadamardProduct");

		MultipleInputs.addInputPath(conf, input_tensor_path, TextInputFormat.class, NModeMatrixHadamardProductTensorMapper.class);
		MultipleInputs.addInputPath(conf, input_factor_path, TextInputFormat.class, NModeMatrixHadamardProductFactorMapper.class);

		fs.delete(output_path, true);
		conf.setReducerClass(NModeMatrixHadamardProduct.NModeMatrixHadamardProductReducer.class);
		FileOutputFormat.setOutputPath(conf, output_path);  

		conf.setNumReduceTasks( nreducers );

		conf.setOutputKeyClass(LongWritable.class);
		conf.setOutputValueClass(Text.class);

		return conf;
	}

	// Configure configSuccessiveNModeMatrixHadamardProduct
	// input:
	// 	- current_mode: current mode
	// 	- out_path: output path
	// output: result of n-mode MatrixHadamardProduct
	protected JobConf configSuccessiveNModeMatrixHadamardProduct (Path prev_tensor_path, Path input_path, int product_mode, Path output_path) throws Exception {
		final JobConf conf = new JobConf(getConf(), CMTFtest.class);

		conf.set("product_mode", "" + product_mode);
		conf.set("rank", "" + rank);
		conf.setJobName("[Iteration "+ iteration +"] product_mode="+product_mode+" configSuccessiveNModeMatrixHadamardProduct");

		MultipleInputs.addInputPath(conf, prev_tensor_path, TextInputFormat.class, SuccessiveNModeMatrixHadamardProductTensorMapper.class);
		MultipleInputs.addInputPath(conf, input_path, TextInputFormat.class, SuccessiveNModeMatrixHadamardProductFactorMapper.class);

		fs.delete(output_path, true);
		conf.setReducerClass(SuccessiveNModeMatrixHadamardProductReducer.class);
		FileOutputFormat.setOutputPath(conf, output_path);  

		conf.setNumReduceTasks( nreducers );

		conf.setOutputKeyClass(LongWritable.class);
		conf.setOutputValueClass(Text.class);

		return conf;
	}

	// configCollapse: Configure MultipleCollapse
	// input:
	// 	- in_path: paths of input (n-mode Hadamard product)
	// 	- current_mode: current mode
	// 	- output_path: output path
	// output: result of mode product
	protected JobConf configCollapse (Path in_path, int current_mode, Path output_path) throws Exception {
		final JobConf conf = new JobConf(getConf(), CMTFtest.class);

		conf.set("current_mode", "" + current_mode);
		conf.set("rank", "" + rank);
		conf.setJobName("[Iteration "+ iteration +"] Mode "+current_mode+" Collapse");

		MultipleInputs.addInputPath(conf, in_path, TextInputFormat.class, Collapse.CollapseMultipleModeMapper.class);
		conf.setReducerClass(Collapse.CollapseReducer.class);   

		fs.delete(output_path, true);

		FileOutputFormat.setOutputPath(conf, output_path);  
		conf.setNumReduceTasks( nreducers );
		conf.setOutputKeyClass(LongWritable.class);
		conf.setOutputValueClass(Text.class);

		return conf;
	}

	// configCollapse: Configure MultipleCollapse
	// input:
	// 	- in_path: paths of input (n-mode Hadamard product)
	// 	- current_mode: current mode
	// 	- output_path: output path
	// output: result of mode product
	protected JobConf configMultipleCollapse (ArrayList<Path> in_path, int current_mode, Path output_path) throws Exception {
		final JobConf conf = new JobConf(getConf(), CMTFtest.class);

		conf.set("current_mode", "" + current_mode);
		conf.set("rank", "" + rank);
		conf.setJobName("[Iteration "+ iteration +"] Mode "+current_mode+" Collapse");

		MultipleInputs.addInputPath(conf, in_path.get(0), TextInputFormat.class, CollapseMultipleModeMapper.class);
		if(in_path.size() > 1) MultipleInputs.addInputPath(conf, in_path.get(1), TextInputFormat.class, Collapse.CollapseMatrixMapper.class);

		conf.setReducerClass(Collapse.CollapseReducer.class);   

		fs.delete(output_path, true);

		FileOutputFormat.setOutputPath(conf, output_path);  
		conf.setNumReduceTasks( nreducers );
		conf.setOutputKeyClass(LongWritable.class);
		conf.setOutputValueClass(Text.class);

		return conf;
	}

	// Configure ConvertRowwise 
	// input:
	// 	- in_path, output: input and output path
	// 	- R: the number of column
	// 	- factor: indicator of factor
	// 	- nreducers: the number of reducers
	// output: rowwise factor
	protected JobConf configConvertRowwise (String in_path, Path out_path, int current_mode) throws Exception
	{
		final JobConf conf = new JobConf(getConf(), CMTFtest.class);
		conf.set("R", "" + rank);
		conf.set("current_mode", "" + current_mode);
		conf.setJobName("[Iteration "+ iteration +"] ConvertRowwise (mode: " + current_mode + ")");

		conf.setMapperClass(ConvertRowwiseMapper.class);        
		conf.setReducerClass(ConvertRowwiseReducer.class);        

		final FileSystem fs = FileSystem.get(conf);
		fs.delete(out_path, true);

		FileInputFormat.setInputPaths(conf, new Path(in_path));
		FileOutputFormat.setOutputPath(conf, out_path);  
		conf.setNumReduceTasks( nreducers );
		conf.setOutputKeyClass(LongWritable.class);
		conf.setOutputValueClass(Text.class);

		return conf;
	}

	// Configure ConvertColumnwise
	// input:
	// 	- in_path, output: input and output path
	// 	- R: the number of column
	// 	- factor: indicator of factor
	// 	- nreducers: the number of reducers
	// output: rowwise factor
	protected JobConf configConvertColumnwise (Path in_path, Path out_path, int current_mode) throws Exception
	{
		final JobConf conf = new JobConf(getConf(), CMTFtest.class);
		conf.set("current_mode", "" + current_mode);
		conf.setJobName("[Iteration "+ iteration +"] ConvertColumnwise (mode: " + current_mode + ")");

		conf.setMapperClass(ConvertColumnwise.ConvertColumnwiseMapper.class);              

		final FileSystem fs = FileSystem.get(conf);
		fs.delete(out_path, true);

		FileInputFormat.setInputPaths(conf, in_path);
		FileOutputFormat.setOutputPath(conf, out_path);  
		conf.setNumReduceTasks( 0 );
		conf.setOutputKeyClass(LongWritable.class);
		conf.setOutputValueClass(Text.class);

		return conf;
	}

	// Configure MatrixTranspose
	// input:
	// 	- in_path, output: input and output path
	// output: rowwise factor
	protected JobConf configMatrixTranspose (Path in_path, Path out_path) throws Exception
	{
		final JobConf conf = new JobConf(getConf(), CMTFtest.class);
		conf.setJobName("[Iteration "+ iteration +"] MatrixTranspose");

		conf.setMapperClass(MatrixTranspose.MatrixTransposeMapper.class);              

		final FileSystem fs = FileSystem.get(conf);
		fs.delete(out_path, true);

		FileInputFormat.setInputPaths(conf, in_path);
		FileOutputFormat.setOutputPath(conf, out_path);  
		conf.setNumReduceTasks( 0 );
		conf.setOutputKeyClass(Text.class);
		conf.setOutputValueClass(Text.class);

		return conf;
	}

	// Configure MTM
	// input:
	// 	- in_path, out_path: input and output path
	// 	- R: row and column dimensionality (M: R by R matrix)
	// 	- nreducers: the number of reducers 
	// output: result of M^T * M
	protected JobConf configMTM (String in_path_array, int current_mode, Path out_path) throws Exception
	{
		final JobConf conf = new JobConf(getConf(), SCouT.class);
		conf.set("nreducers", "" + nreducers);
		conf.set("R", "" + rank);
		conf.setJobName("[Iteration "+ iteration +"] MTM");

		conf.setMapperClass(MTMMapper.class);        
		conf.setReducerClass(MTMReducer.class);                
		conf.setNumMapTasks(nreducers);
		final FileSystem fs = FileSystem.get(conf);
		fs.delete(out_path, true);

		for(int i=1; i<=dimensionsTensor.size(); i++) {
			if(i != current_mode) {
				FileInputFormat.addInputPath(conf, new Path(in_path_array + i));  
			}
		}

		FileOutputFormat.setOutputPath(conf, out_path);  
		conf.setNumReduceTasks( 1 );

		conf.setOutputKeyClass(LongWritable.class);
		conf.setOutputValueClass(Text.class);

		return conf;
	}

	// Configure MTM
	// input:
	// 	- in_path, out_path: input and output path
	// 	- R: row and column dimensionality (M: R by R matrix)
	// 	- nreducers: the number of reducers 
	// output: result of M^T * M
	protected JobConf configMTM (Path in_path, Path out_path) throws Exception
	{
		final JobConf conf = new JobConf(getConf(), CMTFtest.class);
		conf.set("nreducers", "" + nreducers);
		conf.set("R", "" + rank);
		conf.setJobName("[Iteration "+ iteration +"] MTM");

		conf.setMapperClass(MTMMapper.class);        
		conf.setReducerClass(MTMReducer.class);        
		conf.setNumMapTasks(nreducers);
		final FileSystem fs = FileSystem.get(conf);
		fs.delete(out_path, true);

		FileInputFormat.setInputPaths(conf, in_path);  
		FileOutputFormat.setOutputPath(conf, out_path);  
		conf.setNumReduceTasks( 1 );

		conf.setOutputKeyClass(LongWritable.class);
		conf.setOutputValueClass(Text.class);

		return conf;
	}
}
