package SCouT;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;

import org.apache.hadoop.io.DoubleWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.MapReduceBase;
import org.apache.hadoop.mapred.Mapper;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.hadoop.mapred.Reducer;
import org.apache.hadoop.mapred.Reporter;

public class NModeMatrixHadamardProduct {
	/**
	 * 
	 * n-mode matrix Hadamard Product mapper : (X *_n A(n))
	 * input: (if current mode == n)
	 * 	- tensor: I_1, I_2, ..., I_N, tensor_value
	 * 	- matrix: I_n, r_1 matrix_value, ..., r_R matrix_value
	 * 
	 * output: (if current mode == n)
	 * 	- tensor: key(I_n), value(I_1, I_2, ..., I_N, tensor_value)
	 *  - factor: key(I_n), value(r_1 matrix_value, ..., r_R matrix_value)
	 * 
	 */
	public static class NModeMatrixHadamardProductTensorMapper extends MapReduceBase implements Mapper<LongWritable, Text, LongWritable, Text> {
		int product_mode = 0;
		int rank = 0;
		public void configure(JobConf job) {
			product_mode = Integer.parseInt(job.get("product_mode"));
			rank = Integer.parseInt(job.get("rank"));
			System.out.println("NModeMatrixHadamardProductTensorMapper : product_mode = " + product_mode + " rank = " + rank);
		}

		public void map (LongWritable key, Text value, OutputCollector<LongWritable, Text> output, Reporter reporter) throws IOException {
			String line_text = value.toString(); // tensor: I_1, I_2, ..., I_N, tensor_value
			final String[] tokens = line_text.split("\t");

			LongWritable out_key = new LongWritable(Long.parseLong( tokens[product_mode-1]));
			Text out_value = new Text();
			String out_str = "" + tokens[0];

			for(int i=1; i<tokens.length-1; i++) {
				out_str += "\t" + tokens[i];
			}
			out_str += ":" + tokens[tokens.length-1];

			out_value.set(out_str);
			output.collect(out_key, out_value); // key: product_mode_index, value: I_1, I_2, ..., I_N:tensor_value
		}
	}

	public static class NModeMatrixHadamardProductFactorMapper extends MapReduceBase implements Mapper<LongWritable, Text, LongWritable, Text> {
		int product_mode = 0;
		int rank = 0;
		public void configure(JobConf job) {
			product_mode = Integer.parseInt(job.get("product_mode"));
			rank = Integer.parseInt(job.get("rank"));
			System.out.println("NModeMatrixHadamardProductFactorMapper : product_mode = " + product_mode + " rank = " + rank);
		}

		public void map (LongWritable key, Text value, OutputCollector<LongWritable, Text> output, Reporter reporter) throws IOException {
			String line_text = value.toString(); // matrix: I_n, r matrix_value
			final String[] tokens = line_text.split("\t");
			long row = Long.parseLong(tokens[0]);
			output.collect( new LongWritable(row), new Text("f" + line_text)); // key: row, value: frow column value, ..., column value
		}
	}

	/**
	 * 
	 * n-mode matrix Hadamard Product reducer : (X *_n A(n))
	 * input: (if current mode == n)
	 * 	- tensor: key(I_n), value(I_1, ...,I_N, tensor_value)
	 * 	- factor: key(I_n), value(column, factor_value)
	 * output: (if current mode == n)
	 *	- (I_1, ..., I_N, column, tensor_value * factor_value)
	 *
	 */
	public static class NModeMatrixHadamardProductReducer extends MapReduceBase	implements Reducer<LongWritable, Text, Text, DoubleWritable> {
		int product_mode = 0;
		int rank = 0;
		public void configure(JobConf job) {
			product_mode = Integer.parseInt(job.get("product_mode"));
			rank = Integer.parseInt(job.get("rank"));
			System.out.println("NModeMatrixHadamardProductReducer : product_mode = " + product_mode + " rank = " + rank);
		}

		public void reduce (final LongWritable key, final Iterator<Text> values, final OutputCollector<Text, DoubleWritable> output, final Reporter reporter) throws IOException {
			ArrayList<String> tensor_value_list = new ArrayList<String>();
			Double factor_value_list[] = new Double[rank];
			boolean getFactor = false;
			Text out_key = new Text();
			DoubleWritable out_value = new DoubleWritable();

			while (values.hasNext()) {
				String line_text = values.next().toString();
				if(line_text.charAt(0) == 'f') { // factor
					final String[] tokens = line_text.split("\t");
					for(int j=1; j<tokens.length; j+=2) {
						factor_value_list[j/2] = Double.parseDouble(tokens[j+1]);
					}
					for(int i=0; i<tensor_value_list.size(); i++) {
						for(int j=0; j<rank; j++) {
							String[] tensor_str = tensor_value_list.get(i).split(":");
							out_key.set(tensor_str[0] + "\t" + (j+1));
							out_value.set(Double.parseDouble(tensor_str[1])* factor_value_list[j] );
							output.collect( out_key , out_value);
						}
					}
					getFactor = true;
					tensor_value_list.clear();
				} else { // tensor 
					if(getFactor == false) {
						tensor_value_list.add(line_text);
					} else {
						for(int j=0; j<rank; j++) {
							String[] tensor_str = line_text.split(":");
							out_key.set(tensor_str[0] + "\t" + (j+1));
							out_value.set(Double.parseDouble(tensor_str[1])* factor_value_list[j] );
							output.collect( out_key , out_value);
						}
					}
				}
			}
		}
	}
}

